/*
* SACD Decoder plugin
* Copyright (c) 2011-2016 Maxim V.Anisiutkin <maxim.anisiutkin@gmail.com>
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with FFmpeg; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
*/

#include "dsd_stream_service_impl.h"

dsd_stream_service_impl::dsd_stream_service_impl() {
	InitializeCriticalSection(&m_chunk_cs);
	m_chunk_read_samples = 0;
	m_samples = 0;
	m_duration = m_max_duration = 0.0;
}

dsd_stream_service_impl::~dsd_stream_service_impl() {
	flush();
	DeleteCriticalSection(&m_chunk_cs);
}

t_size dsd_stream_service_impl::get_sample_count() {
	return m_samples;
}

double dsd_stream_service_impl::get_duration() {
	return m_duration;
}

void dsd_stream_service_impl::set_max_duration(double p_duration) {
	m_max_duration = p_duration;
}

double dsd_stream_service_impl::get_watermark() {
	return (m_max_duration > 0.0) ? m_duration / m_max_duration : 0.0;
}

t_samplespec dsd_stream_service_impl::get_spec() {
	t_samplespec spec;
	EnterCriticalSection(&m_chunk_cs);
	if (m_chunks.get_count() > 0) {
		pfc::iterator<dsd_chunk_t*> chunk_iter = m_chunks.first();
		spec = (*chunk_iter)->get_spec();
	}
	LeaveCriticalSection(&m_chunk_cs);
	return spec;
}

bool dsd_stream_service_impl::is_spec_change() {
	return m_last_read_spec != get_spec();
}

void dsd_stream_service_impl::reset_spec_change() {
	m_last_read_spec = get_spec();
}

dsd_chunk_t* dsd_stream_service_impl::get_first_chunk_ptr() {
	dsd_chunk_t* chunk_ptr = nullptr;
	EnterCriticalSection(&m_chunk_cs);
	if (m_chunks.get_count() > 0) {
		chunk_ptr = *(m_chunks.first());
	}
	LeaveCriticalSection(&m_chunk_cs);
	return chunk_ptr;
}

void dsd_stream_service_impl::remove_first_chunk() {
	EnterCriticalSection(&m_chunk_cs);
	if (m_chunks.get_count() > 0) {
		pfc::iterator<dsd_chunk_t*> chunk_iter = m_chunks.first();
		(*chunk_iter)->get_duration();
		m_samples -= (*chunk_iter)->get_sample_count();
		m_duration -= (*chunk_iter)->get_duration();
		delete *chunk_iter;
		m_chunks.remove(chunk_iter);
	}
	LeaveCriticalSection(&m_chunk_cs);
}

t_size dsd_stream_service_impl::read(t_uint8* p_buffer, t_size p_samples) {
	t_size read_samples = 0;
	EnterCriticalSection(&m_chunk_cs);
	while (m_chunks.get_count() > 0) {
		pfc::iterator<dsd_chunk_t*> chunk_iter = m_chunks.first();
		t_samplespec chunk_spec = (*chunk_iter)->get_spec();
		if (chunk_spec != m_last_read_spec) {
			break;
		}
		t_uint8* chunk_data = (*chunk_iter)->get_data();
		t_size chunk_samples = (*chunk_iter)->get_sample_count();
		unsigned chunk_channels = (*chunk_iter)->get_channels();
		double sample_duration = (*chunk_iter)->get_duration() / (*chunk_iter)->get_sample_count();
		t_size samples_to_read = p_samples - read_samples;
		if (chunk_samples - m_chunk_read_samples > samples_to_read) {
			memcpy(p_buffer + read_samples * chunk_channels, chunk_data + m_chunk_read_samples * chunk_channels, samples_to_read * chunk_channels);
			m_chunk_read_samples += samples_to_read;
			read_samples += samples_to_read;
			m_samples -= samples_to_read;
			m_duration -= sample_duration * samples_to_read;
		}
		else {
			memcpy(p_buffer + read_samples * chunk_channels, chunk_data + m_chunk_read_samples * chunk_channels , (chunk_samples - m_chunk_read_samples) * chunk_channels);
			delete *chunk_iter;
			m_chunks.remove(chunk_iter);
			read_samples += chunk_samples - m_chunk_read_samples;
			m_samples -= chunk_samples - m_chunk_read_samples;
			m_duration -= sample_duration * (chunk_samples - m_chunk_read_samples);
			m_chunk_read_samples = 0;
		}
		if (p_samples <= read_samples) {
			break;
		}
	}
	LeaveCriticalSection(&m_chunk_cs);
	return read_samples;
}

void dsd_stream_service_impl::write(const t_uint8* p_buffer, t_size p_samples, t_size p_channels, unsigned p_samplerate, unsigned p_channel_config) {
	dsd_chunk_t* chunk = new dsd_chunk_t(p_buffer, p_samples, p_channels, p_samplerate, p_channel_config);
	EnterCriticalSection(&m_chunk_cs);
	m_chunks.add_item(chunk);
	m_samples += chunk->get_sample_count();
	m_duration += chunk->get_duration();
	LeaveCriticalSection(&m_chunk_cs);
}

void dsd_stream_service_impl::flush() {
	EnterCriticalSection(&m_chunk_cs);
	while (m_chunks.get_count() > 0) {
		pfc::iterator<dsd_chunk_t*> chunk_iter = m_chunks.first();
		delete *chunk_iter;
		m_chunks.remove(chunk_iter);
	}
	m_chunk_read_samples = 0;
	m_samples = 0;
	m_duration = 0.0;
	LeaveCriticalSection(&m_chunk_cs);
}

service_factory_single_t<dsd_stream_service_impl> g_dsd_stream_service_factory;
